import unittest
from unittest import mock
import time
import logging
import os
import whalealert.settings as settings
from whalealert.publisher.writer import Writer
from whalealert.whalealert import WhaleAlert

logging.disable(logging.CRITICAL)
TEST_WORKING_DIR = os.path.join('tests', 'test_working_directory')
successful_call_with_trans = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 200,
    'error_message': '',
    'transaction_count': 3
}
successful_call_no_trans = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 200,
    'error_message': '',
    'transaction_count': 0
}
failed_call = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 401,
    'error_message': 'Bad API Key',
    'transaction_count': 0
}
bad_status = {'timestamp': '2020-05-04T16:20:37.276101'}
test_extra_key = [{
    'blockchain': 'bitcoin',
    'symbol': 'btc',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1,
    'extra_key': 1
}]

test_bad_key = [{
    'blockchain': 'bitcoin',
    'symbol': 'btc',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from_a_bad_key': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1,
    'extra_key': 1
}]

test_good_data = [{
    'blockchain': 'bitcoin',
    'symbol': 'btc',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1
}, {
    'blockchain': 'ethereum',
    'symbol': 'usdt',
    'id': '662472178',
    'transaction_type': 'transfer',
    'hash': '246fc938c61f3ce1c59ede4f4aca302f46a9f894a300313d8f6c6600613f8659',
    'from': {
        'address': 'c3df5dfd8c8806961be9d02bd0832ec2a01e3413',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 500000,
    'amount_usd': 503153.03,
    'transaction_count': 1
}, {
    'blockchain': 'tron',
    'symbol': 'usdt',
    'id': '662472729',
    'transaction_type': 'transfer',
    'hash': '55f5fa53fd9c1a8430d79248e45f130b6ac08892ecd68e54cc600228919058d0',
    'from': {
        'address': 'c5a8859c44ac8aa2169afacf45b87c08593bec10',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': '7286c758578a2457e9ba36d46893176582acefb5',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874451,
    'amount': 500000,
    'amount_usd': 500000,
    'transaction_count': 1
}, {
    'blockchain': 'neo',
    'symbol': 'neo',
    'id': '662472747',
    'transaction_type': 'transfer',
    'hash': '40d261b3c94b5c174c802e3d452bd3ad36a3ba6171cbae670d89137ec7a6932e',
    'from': {
        'address': 'b5cebb87ff1a48ca28abc25b4584ff5148d1808e',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': 'f4561c710ba450aab302ffc3557ee59bbce94ca6',
        'owner_type': 'unknown',
        'owner': ''
    },
    'timestamp': 1588874451,
    'amount': 1691976.5,
    'amount_usd': 1689525,
    'transaction_count': 1
}, {
    'blockchain': 'ethereum',
    'symbol': 'usdc',
    'id': '662472982',
    'transaction_type': 'transfer',
    'hash': '866bd80566a494ac83da9822110fdb81189eefc4024208becfdafd3d76f8d921',
    'from': {
        'address': '8cce0f304933a784026f85a14daadb0a33ab06cf',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': '159d563ee3928cfa2bd3631b39266271888fe882',
        'owner_type': 'unknown',
        'owner': ''
    },
    'timestamp': 1588874481,
    'amount': 500000,
    'amount_usd': 500000,
    'transaction_count': 1
}]


def cleanup_working_directories():
    try:
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.database_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.status_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.input_configuation_filename))
        os.removedirs(os.path.join(TEST_WORKING_DIR, settings.data_file_directory))
        os.removedirs(TEST_WORKING_DIR)
    except Exception as e_r:
        _ = e_r


class WritingCurrentAndAllTimeStatus(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())

    def tearDown(self):
        cleanup_working_directories()

    def test_writing_status_with_no_working_directory_handled_ok(self):
        writer = Writer(self.whale.get_status(), self.whale.get_database())
        writer.write_status(successful_call_no_trans)

    def test_writing_a_successful_call_updates_success_no_trans(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_successful_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_successful_calls)
        success = self.writer.write_status(successful_call_no_trans)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_successful_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_successful_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time + 1)
        self.assertEqual(new_value_current, previous_value_current + 1)

    def test_writing_a_successful_call_updates_success_with_trans(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_successful_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_successful_calls)
        success = self.writer.write_status(successful_call_with_trans)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_successful_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_successful_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time + 1)
        self.assertEqual(new_value_current, previous_value_current + 1)

    def test_writing_a_successful_call_doesnt_change_failed_calls(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_failed_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_failed_calls)
        success = self.writer.write_status(successful_call_no_trans)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_failed_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_failed_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time)
        self.assertEqual(new_value_current, previous_value_current)

    def test_writing_a_successful_call_doesnt_change_failed_calls_with_trans(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_failed_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_failed_calls)
        success = self.writer.write_status(successful_call_with_trans)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_failed_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_failed_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time)
        self.assertEqual(new_value_current, previous_value_current)

    def test_writing_a_failed_call_updates_failed(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_failed_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_failed_calls)
        success = self.writer.write_status(failed_call)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_failed_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_failed_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time + 1)
        self.assertEqual(new_value_current, previous_value_current + 1)

    def test_writing_a_failed_call_doesnt_change_successful_calls(self):
        previous_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                        settings.status_file_option_successful_calls)
        previous_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                       settings.status_file_option_successful_calls)
        success = self.writer.write_status(failed_call)
        self.reload_status()
        new_value_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                   settings.status_file_option_successful_calls)
        new_value_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_successful_calls)
        self.assertEqual(success, True)
        self.assertEqual(new_value_all_time, previous_value_all_time)
        self.assertEqual(new_value_current, previous_value_current)

    def test_writing_failed_and_successfull_calls_calculates_percent(self):
        self.writer.get_status_object().set_value(settings.status_file_all_time_section_name,
                                                  settings.status_file_option_success_rate, 0)
        self.writer.get_status_object().set_value(settings.status_file_current_session_section_name,
                                                  settings.status_file_option_success_rate, 0)
        self.writer.write_status(successful_call_no_trans)
        self.writer.write_status(successful_call_no_trans)
        self.writer.write_status(successful_call_no_trans)
        self.writer.write_status(failed_call)
        success = self.writer.write_status(failed_call)
        self.reload_status()
        percent_all_time = self.get_status_value(settings.status_file_all_time_section_name,
                                                 settings.status_file_option_success_rate)
        percent_current = self.get_status_value(settings.status_file_current_session_section_name,
                                                settings.status_file_option_success_rate)
        self.assertEqual(percent_all_time, 60.00)
        self.assertEqual(percent_current, 60.00)
        self.assertEqual(success, True)

    def test_writing_health(self):

        health_length = settings.health_list_length
        self.assertGreater(health_length, 10)
        self.writer.write_status(failed_call)
        success = self.writer.write_status(failed_call)
        expected_health = round(100 * (health_length - 2) / health_length, 1)
        self.reload_status()
        current_health = self.get_status_value(settings.status_file_current_session_section_name,
                                               settings.status_file_option_health)
        self.assertEqual(current_health, expected_health)
        self.assertEqual(success, True)

    def reload_status(self):
        target_file = os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.status_file_name)
        self.writer.get_status_object().set_configuration_file(target_file)

    def get_status_value(self, section, key):
        return self.writer.get_status_object().get_value(section, key)


class WrintingLastCallStats(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())

    def tearDown(self):
        cleanup_working_directories()

    def reload_status(self):
        target_file = os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.status_file_name)
        self.writer.get_status_object().set_configuration_file(target_file)

    def get_status_value(self, section, key):
        return self.writer.get_status_object().get_value(section, key)

    def test_successful_call_writen_ok(self):
        status = successful_call_with_trans
        success = self.writer.write_status(status)
        self.reload_status()
        self.assertEqual(
            self.get_status_value(settings.status_file_last_good_call_section_name,
                                  settings.status_file_option_timeStamp), successful_call_with_trans['timestamp'])
        self.assertEqual(
            self.get_status_value(settings.status_file_last_good_call_section_name,
                                  settings.status_file_option_transaction_count),
            successful_call_with_trans['transaction_count'])
        self.assertEqual(success, True)

    def test_error_call_writen_ok(self):
        status = failed_call
        success = self.writer.write_status(status)
        self.reload_status()
        self.assertEqual(
            self.get_status_value(settings.status_file_last_failed_secion_name, settings.status_file_option_timeStamp),
            failed_call['timestamp'])
        self.assertEqual(
            self.get_status_value(settings.status_file_last_failed_secion_name, settings.status_file_option_error_code),
            failed_call['error_code'])
        self.assertEqual(
            self.get_status_value(settings.status_file_last_failed_secion_name,
                                  settings.status_file_option_error_message), failed_call['error_message'])
        self.assertEqual(success, True)


class HandlingErrors(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())

    def tearDown(self):
        cleanup_working_directories()

    def test_status_key_error(self):
        success = self.writer.write_status(bad_status)
        self.assertIs(success, False)

    def test_status_handle_missing_status_file(self):
        cleanup_working_directories()
        success = self.writer.write_status(successful_call_with_trans)
        self.assertIs(success, False)

    def test_handle_bad_type(self):
        writer = Writer(None, None)
        success = writer.write_status(successful_call_with_trans)
        self.assertIs(success, False)


class WritingData(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())

    def tearDown(self):
        cleanup_working_directories()

    def test_writing_new_data_adds_tables_to_database(self):
        transactions = test_good_data
        success = self.writer.write_transactions(transactions)
        database = self.writer.get_database()
        tables = database.get_table_names()
        expected_tables = ['bitcoin', 'ethereum', 'tron', 'neo']
        self.assertEqual(success, True)
        self.assertEqual(tables, expected_tables)

    def test_writing_new_data_adds_correct_columns_to_table(self):
        transactions = test_good_data
        success = self.writer.write_transactions(transactions)
        database = self.writer.get_database()
        columns = database.get_column_names('bitcoin')
        expected_columns = sorted(list(settings.database_columns.keys()))
        self.assertEqual(success, True)
        self.assertEqual(expected_columns, columns)

    def test_writing_new_data_last_time_entry_is_correct(self):
        transactions = test_good_data
        success = self.writer.write_transactions(transactions)
        database = self.writer.get_database()
        last_entry = database.get_last_time_entry('bitcoin')
        expected_entry = {
            'blockchain': 'bitcoin',
            'symbol': 'btc',
            'id': '662472177',
            'transaction_type': 'transfer',
            'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
            'from_address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
            'from_owner': 'binance',
            'from_owner_type': 'exchange',
            'to_address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
            'to_owner': 'binance',
            'to_owner_type': 'exchange',
            'timestamp': 1588874414,
            'amount': 3486673,
            'amount_usd': 3508660.2,
            'transaction_count': 1
        }
        self.assertEqual(success, True)
        self.assertEqual(expected_entry, last_entry)

    def test_writing_new_data_extra_key_returns_false(self):
        transaction = test_extra_key
        success = self.writer.write_transactions(transaction)
        self.assertEqual(success, False)

    def test_writing_empty_transactions_returns_false(self):
        transactions = []
        success = self.writer.write_transactions(transactions)
        self.assertEqual(success, False)

    def test_writing_with_a_bad_key_returns_false(self):
        transactions = test_bad_key
        success = self.writer.write_transactions(transactions)
        self.assertEqual(success, False)

    def test_writing_with_a_bad_dictionary_returns_false(self):
        transactions = dict(test_bad_key[0])
        success = self.writer.write_transactions(transactions)
        self.assertEqual(success, False)

    def test_writing_successive_entries(self):
        transactions = test_good_data
        success = self.writer.write_transactions(transactions)
        success = self.writer.write_transactions(transactions)
        success = self.writer.write_transactions(transactions)
        database = self.writer.get_database()
        df = database.table_to_df('bitcoin')
        self.assertEqual(len(df), 3)

        # There are two ethereum transaction in the good data
        df = database.table_to_df('ethereum')
        self.assertEqual(len(df), 6)

        self.assertEqual(success, True)

    def test_writing_wrong_data_type(self):
        transactions = 'A string'
        success = self.writer.write_transactions(transactions)
        self.assertEqual(success, False)



"""
API wrapper for getting transactions from Whale Alert API
"""

import logging
import json
import time
import datetime
from requests import Session
from requests.exceptions import Timeout, TooManyRedirects
import whalealert.settings as settings

log = logging.getLogger(__name__)


class Transactions():
    """
    API wrapper for getting transactions from Whale Alert API
    """
    def __init__(self):
        self.__session = Session()
        self.__last_timestamp = int(time.time())
        self.__last_cursor = None
        self.__call_return_time = int(time.time())
        logging.debug("Transactions object created")

    def get_last_timestamp(self):
        """ Get the timestamp of the last successful call

        Returns:
        Unix timestamp (int): Last successful call time.
        """
        return self.__last_timestamp

    def get_last_cursor(self):
        """ Get the API cursor for the last successful call.

        Returns:
        cursor (string): Last successful API call cursor provided by Whale Alert
        """
        return self.__last_cursor

    def get_transactions(self, start_time, end_time, api_key, cursor, min_value, limit):
        """ Make an API call to Whale Alert to get the latest transactions.

        The free API has the following limitions:
        - Historical limit of 3600 seconds. (start_time)
        - Minimum transactino value is 500000 USD
        - Limit of 100 transactions

        Parameters:
        start_time (int): Unix timestamp from which to start listing transactions. (exclusive)
        end_time (int): Unix timestamp from where to stop listing transactions (inclusive). Use None to get to current time.
        api_key (str): The API key to use for the transaction.
        cursor (str): The pagnation cursor from a previous transaction. Use None to ignore this parameter.
        min_value (int): The minimum value of transaction to return.
        limit (int): The maximum number of transactions to return (maximum = 100)

        Returns:
        success (bool) : If true, then a successful call was made (Return code of 200). Return false otherwise.
        transactions (list or None):
        - success = True: A list containing a dictionary for each transaction.
        - success = False: None
        status (dict): A dictionary containing the timestamp, error_code and error_message for the transaction.
        """
        parameters = self.__form_request(start_time, end_time, api_key, cursor, min_value, limit)
        log.debug("Attempting API call at '{}' with parameters '{}'".format(settings.whale_get_transactions_url,
                                                                            parameters))

        success, response, status = self.__attempt_call(parameters)
        if success is not True:
            return success, None, status

        success, transactions, status = self.__check_response(response)
        return success, transactions, status

    def __attempt_call(self, parameters):
        retries = settings.whale_retries_on_failure
        attempt = 1
        while retries >= 0:
            try:
                response = self.__session.get(settings.whale_get_transactions_url,
                                              params=parameters,
                                              timeout=settings.whale_call_timeout_seconds)
                self.__call_return_time = int(time.time())
                retries = -1
            except ConnectionError:
                retries, attempt, done = self.__evalulate_attempt(retries, attempt)
                if not done:
                    continue
                status = self.__make_custom_status(1, "Internal error: Connection exception when conduction API call",
                                                   0)
                return False, None, status
            except Timeout:
                retries, attempt, done = self.__evalulate_attempt(retries, attempt)
                if not done:
                    continue
                status = self.__make_custom_status(2, "Internal error: Timeout exception when conduction API call", 0)
                return False, None, status
            except TooManyRedirects:
                retries, attempt, done = self.__evalulate_attempt(retries, attempt)
                if not done:
                    continue
                status = self.__make_custom_status(
                    3, "Internal error: TooManyRedirect exception when conduction API call", 0)
                return False, None, status
            except Exception as e_r:
                retries, attempt, done = self.__evalulate_attempt(retries, attempt)
                if not done:
                    continue
                status = self.__make_custom_status(4,
                                                   "Internal error: Exception {} when conduction API call".format(e_r),
                                                   0)
                return False, None, status
        return True, response, None

    def __evalulate_attempt(self, retries, attempt):
        if retries >= 1:
            retries = retries - 1
            time.sleep(2 * attempt)
            log.warning("Exception occurred on API call attempt {} of {}".format(attempt,
                                                                                 settings.whale_retries_on_failure))
            attempt = attempt + 1
            done = False
        else:
            done = True
        return retries, attempt, done

    def __check_response(self, response):
        try:
            json_fields = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            status = self.__make_custom_status(5, "Internal error: Error parsing JSON object from received response", 0)
            return False, None, status
        except Exception as e:
            status = self.__make_custom_status(
                6,
                "Internal error: Exception {} when parsing JSON object from received response. Response =  {}".format(
                    e, response.text), 0)
            return False, None, status

        if response.status_code != 200:
            status = self.__parse_error_response(json_fields, response.status_code)
            return False, None, status

        success, transactions, status = self.__parse_good_response(json_fields)
        return success, transactions, status

    def __parse_good_response(self, json_fields):
        good_keys, count, status = self.__check_main_keys(json_fields)
        if good_keys is False:
            return False, None, status

        if count == 0:
            log.info("Successful API call returned {} transactions".format(count))
            return True, [], self.__make_custom_status(200, '', 0)

        good_keys, transactions, status = self.__check_transactions_keys(json_fields)
        if good_keys is False:
            return False, None, status

        self.__last_cursor = json_fields[settings.whale_success_cursor]
        self.__last_timestamp = self.__call_return_time
        log.info("Successful API call returned {} transactions".format(len(transactions)))
        return True, transactions, self.__make_custom_status(200, '', count)

    def __check_transactions_keys(self, json_fields):
        for transaction in json_fields[settings.whale_success_transactions]:
            try:
                dummy = transaction[settings.whale_transaction_blockchain]
                dummy = transaction[settings.whale_transaction_symbol]
                dummy = transaction[settings.whale_transaction_id]
                dummy = transaction[settings.whale_transaction_transaction_type]
                dummy = transaction[settings.whale_transaction_hash]
                dummy = transaction[settings.whale_transaction_from][settings.whale_transaction_address]
                dummy = transaction[settings.whale_transaction_to][settings.whale_transaction_address]
                dummy = transaction[settings.whale_transaction_timestamp]
                dummy = transaction[settings.whale_transaction_amount]
                dummy = transaction[settings.whale_transaction_amount_usd]
                dummy = transaction[settings.whale_transaction_transaction_count]

                owner_type = transaction[settings.whale_transaction_from][settings.whale_transaction_owner_type]
                transaction[settings.whale_transaction_symbol] = transaction[settings.whale_transaction_symbol].upper()
                if owner_type != 'unknown':
                    dummy = transaction[settings.whale_transaction_from][settings.whale_transaction_owner]
                else:
                    transaction[settings.whale_transaction_from][settings.whale_transaction_owner] = ''

                owner_type = transaction[settings.whale_transaction_to][settings.whale_transaction_owner_type]
                if owner_type != 'unknown':
                    dummy = transaction[settings.whale_transaction_to][settings.whale_transaction_owner]
                else:
                    transaction[settings.whale_transaction_to][settings.whale_transaction_owner] = ''
            except KeyError:
                return False, None, self.__make_custom_status(
                    10, 'Internal error: Error with transactions JSON keys, bad transaction = {}'.format(transaction),
                    0)
        return True, json_fields[settings.whale_success_transactions], None

    def __check_main_keys(self, json_fields):
        try:
            dummy = json_fields[settings.whale_success_result]
            dummy = json_fields[settings.whale_success_cursor]
            count = json_fields[settings.whale_success_count]
            if count > 0:
                transactions = json_fields[settings.whale_success_transactions]
                if len(transactions) != count:
                    status = self.__make_custom_status(
                        8, "Internal error: Transaction count doesn't match reported count. Response = {}".format(
                            json_fields), 0)
                    return False, 0, status
        except KeyError:
            status = self.__make_custom_status(
                9, "Internal error: Problem parsing main keys. Response = {}".format(json_fields), 0)
            return False, 0, status
        return True, count, None

    def __parse_error_response(self, json_response, code):
        try:
            message = json_response[settings.whale_error_message]
        except Exception as e_r:
            code = 7
            message = "Internal error: Cannot read message from error response. Response = {}, Exception = {}".format(
                json_response, e_r)
        return self.__make_custom_status(code, message, 0)

    def __make_custom_status(self, errNo, message, transaction_count):
        custom_status = dict()
        custom_status[settings.status_file_option_timeStamp] = str(datetime.datetime.now().isoformat())
        custom_status[settings.status_file_option_error_code] = errNo
        custom_status[settings.status_file_option_error_message] = message
        custom_status[settings.status_file_option_transaction_count] = transaction_count
        if errNo != 200:
            log.error(message)
        return custom_status

    def __form_request(self, start_time, end_time, api_key, cursor, min_value, limit):
        call_parameters = {}

        # Required for every call
        call_parameters['api_key'] = api_key
        call_parameters['start'] = start_time
        call_parameters['min_value'] = min_value
        call_parameters['limit'] = limit

        if end_time is not None:
            call_parameters['end'] = end_time

        if cursor is not None:
            call_parameters['cursor'] = cursor

        return call_parameters



import unittest
from unittest import mock
import time
import logging
import os
import datetime
import pandas as pd
import whalealert.settings as settings
from whalealert.publisher.writer import Writer
from whalealert.publisher.reader import Reader
from whalealert.whalealert import WhaleAlert

logging.disable(logging.CRITICAL)
TEST_WORKING_DIR = os.path.join('tests', 'test_working_directory')

successful_call_with_trans = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 200,
    'error_message': '',
    'transaction_count': 3
}

test_good_data = [{
    'blockchain': 'bitcoin',
    'symbol': 'BTC',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1188874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1
}, {
    'blockchain': 'ethereum',
    'symbol': 'USDT',
    'id': '662472178',
    'transaction_type': 'transfer',
    'hash': '246fc938c61f3ce1c59ede4f4aca302f46a9f894a300313d8f6c6600613f8659',
    'from': {
        'address': 'c3df5dfd8c8806961be9d02bd0832ec2a01e3413',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1288874414,
    'amount': 500000,
    'amount_usd': 503153.03,
    'transaction_count': 1
}, {
    'blockchain': 'tron',
    'symbol': 'USDT',
    'id': '662472729',
    'transaction_type': 'sold',
    'hash': '55f5fa53fd9c1a8430d79248e45f130b6ac08892ecd68e54cc600228919058d0',
    'from': {
        'address': 'c5a8859c44ac8aa2169afacf45b87c08593bec10',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': '7286c758578a2457e9ba36d46893176582acefb5',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1388874451,
    'amount': 500000000,
    'amount_usd': 500000000.23,
    'transaction_count': 1
}, {
    'blockchain': 'neo',
    'symbol': 'NEO',
    'id': '662472747',
    'transaction_type': 'burn',
    'hash': '40d261b3c94b5c174c802e3d452bd3ad36a3ba6171cbae670d89137ec7a6932e',
    'from': {
        'address': 'b5cebb87ff1a48ca28abc25b4584ff5148d1808e',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': 'f4561c710ba450aab302ffc3557ee59bbce94ca6',
        'owner_type': 'unknown',
        'owner': ''
    },
    'timestamp': 1488874451,
    'amount': 1691976.5,
    'amount_usd': 16895205,
    'transaction_count': 1
}, {
    'blockchain': 'ethereum',
    'symbol': 'ETH',
    'id': '662472982',
    'transaction_type': 'transfer',
    'hash': '866bd80566a494ac83da9822110fdb81189eefc4024208becfdafd3d76f8d921',
    'from': {
        'address': '8cce0f304933a784026f85a14daadb0a33ab06cf',
        'owner_type': 'unknown',
        'owner': ''
    },
    'to': {
        'address': '159d563ee3928cfa2bd3631b39266271888fe882',
        'owner_type': 'unknown',
        'owner': ''
    },
    'timestamp': 1588874481,
    'amount': 500000.1,
    'amount_usd': 500000,
    'transaction_count': 1
}]

expected_pretty_output = '\x1b[33m{} \x1b[0m\x1b[37m500000.10 \x1b[31mETH\x1b[36m (500,000.00 USD) \x1b[0m transferred from \x1b[34munknown\x1b[0m to \x1b[34munknown\x1b[0m.'.format(
    Reader.to_local_time(test_good_data[4]['timestamp']))
expected_pretty_large_sum = '\x1b[33m{} \x1b[0m\x1b[37m1691976.50 \x1b[31mNEO\x1b[36m\x1b[1m (16,895,205.00 USD) \x1b[0m \x1b[31mburned\x1b[0m at \x1b[34munknown\x1b[0m.'.format(
    Reader.to_local_time(test_good_data[3]['timestamp']))
expected_pretty_huge_sum = '\x1b[33m{} \x1b[0m\x1b[37m500000000.00 \x1b[31mUSDT \x1b[36m\x1b[1m\x1b[4m(500,000,000.23 USD)\x1b[0m sold from \x1b[34munknown\x1b[0m to \x1b[34mbinance\x1b[0m.'.format(
    Reader.to_local_time(test_good_data[2]['timestamp']))


def cleanup_working_directories():
    try:
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.database_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.status_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.input_configuation_filename))
        os.removedirs(os.path.join(TEST_WORKING_DIR, settings.data_file_directory))
        os.removedirs(TEST_WORKING_DIR)
    except Exception as e_r:
        _ = e_r


def flattern_transactions(transactions):
    new_trans = []
    for t in transactions:
        new_dict = dict(t)
        new_dict[settings.database_column_from_address] = t['from']['address']
        new_dict[settings.database_column_from_owner] = t['from']['owner']
        new_dict[settings.database_column_from_owner_type] = t['from']['owner_type']
        new_dict[settings.database_column_to_address] = t['to']['address']
        new_dict[settings.database_column_to_owner] = t['to']['owner']
        new_dict[settings.database_column_to_owner_type] = t['to']['owner_type']

        new_dict.pop('from', None)
        new_dict.pop('to', None)
        new_trans.append(new_dict)
    return new_trans


def format_output(data, as_dict=False):
    if data['to']['owner'] == '':
        data['to']['owner'] = 'unknown'
    if data['from']['owner'] == '':
        data['from']['owner'] = 'unknown'
    if as_dict:
        return {
            'timestamp':
            Reader.to_local_time(data['timestamp']),
            'text':
            str("{:.2f}".format(data['amount'])) + " " + data['symbol'].upper() + " (" +
            Reader.format_currency(data['amount_usd']) + " USD) " + "transferred from " + data['from']['owner'] +
            ' to ' + data['to']['owner'] + '.'
        }

    else:
        return Reader.to_local_time(data['timestamp']) + " " + str("{:.2f}".format(
            data['amount'])) + " " + data['symbol'].upper() + " (" + Reader.format_currency(
                data['amount_usd']
            ) + " USD) " + "transferred from " + data['from']['owner'] + ' to ' + data['to']['owner'] + '.\n'


class RequestingStatusByExchange(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())
        self.reader = Reader(self.whale.get_status(), self.whale.get_database())

    def tearDown(self):
        cleanup_working_directories()

    def add_call_to_database(self, data):
        self.writer.write_transactions(data)

    def test_single_blockchain_single_symbol_request_one_result(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['bitcoin']
        request[settings.request_symbols] = ['BTC']
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        data = test_good_data[0]
        expected_output = format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_single_blockchain_multiple_symbols_request_one_result(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        data = test_good_data[4]
        expected_output = format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_single_blockchain_multiple_symbols_request_multiple_results(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 2
        request[settings.request_from_time] = 0
        data = test_good_data[1]
        expected_output = format_output(data)
        data = test_good_data[4]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_multiple_blockchain_multiple_symbols_request_multiple_results(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum', 'tron']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[1]
        expected_output = format_output(data)
        data = test_good_data[2]
        expected_output = expected_output + format_output(data)
        data = test_good_data[4]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_multiple_blockchain_multiple_symbols_request_one_result(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum', 'tron']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        data = test_good_data[4]
        expected_output = format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_invalid_blockchain_is_ignored(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum', 'not_a_blockchain']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[1]
        expected_output = format_output(data)
        data = test_good_data[4]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_invalid_symbols_are_ignored(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum', 'tron']
        request[settings.request_symbols] = ["ETH"]
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[4]
        expected_output = format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_values_only_taken_after_timestamp(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum', 'tron']
        request[settings.request_symbols] = ['USDT', "ETH"]
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 1488874451
        data = test_good_data[4]
        expected_output = format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_blockchain_wildcard_gets_all_blockchain(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['USDT']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[1]
        expected_output = format_output(data)
        data = test_good_data[2]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_symbol_widcard_gets_all_symbols(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['ethereum']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[1]
        expected_output = format_output(data)
        data = test_good_data[4]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_both_wildcards_return_all_results(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        data = test_good_data[0]
        expected_output = format_output(data)
        data = test_good_data[1]
        expected_output = expected_output + format_output(data)
        data = test_good_data[2]
        expected_output = expected_output + format_output(data)
        data = test_good_data[3]
        expected_output = expected_output + format_output(data)
        data = test_good_data[4]
        expected_output = expected_output + format_output(data)
        expected_output = expected_output[:-1]
        output = self.reader.data_request(request)
        self.assertEqual(output, expected_output)

    def test_both_wildcards_return_all_results_as_dict(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        data = test_good_data[4]
        expected_output = format_output(data, as_dict=True)
        output = self.reader.data_request(request, as_dict=True)
        self.assertEqual(output, [expected_output])

    def test_single_result_pretty_format(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        output = self.reader.data_request(request, pretty=True)
        self.assertEqual([output], [expected_pretty_output])

    def test_single_result_pretty_format_medium_sum(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['neo']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        output = self.reader.data_request(request, pretty=True)
        self.assertEqual([output], [expected_pretty_large_sum])

    def test_single_pretty_format_huge_sum(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['tron']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 1
        request[settings.request_from_time] = 0
        output = self.reader.data_request(request, pretty=True)
        self.assertEqual([output], [expected_pretty_huge_sum])

    def test_calling_without_database_returns_empty_string(self):
        reader = Reader()
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        output = reader.data_request(request)
        self.assertEqual(output, '')

    def test_calling_with_empty_database_returns_empty_list(self):
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        output = self.reader.data_request(request, as_dict=True)
        self.assertEqual(output, [])

    def test_calling_with_empty_database_returns_empty_dataFrame(self):
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        output = self.reader.data_request(request, as_df=True)
        self.assertIs(output.empty, True)

    def test_bad_data_request_returns_empty_string(self):
        bad_request = {
            'bad_key': 'sfd',
            'another_bad_key': 'sfd',
        }
        output = self.reader.data_request(bad_request)
        self.assertEqual(output, '')

    def test_getting_all_data_as_df(self):
        self.add_call_to_database(test_good_data)
        request = dict(settings.request_format)
        request[settings.request_blockchain] = ['*']
        request[settings.request_symbols] = ['*']
        request[settings.request_maximum_results] = 5
        request[settings.request_from_time] = 0
        expected_output = pd.DataFrame(flattern_transactions(test_good_data))
        expected_output = expected_output.reindex(sorted(expected_output.columns), axis=1)
        output = self.reader.data_request(request, as_df=True)
        pd.testing.assert_frame_equal(expected_output, output)


class GettingLoggerStatus(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(working_directory=TEST_WORKING_DIR)
        self.writer = Writer(self.whale.get_status(), self.whale.get_database())
        self.reader = Reader(self.whale.get_status(), self.whale.get_database(), self.whale.get_configuration())

    def tearDown(self):
        cleanup_working_directories()

    def write_good_status(self, unix_timestamp):
        status = dict()
        status['timestamp'] = Reader.to_local_time(unix_timestamp)
        status['error_code'] = 200
        status['error_message'] = 'A dud message'
        status['transaction_count'] = 0
        self.writer.write_status(status)

    def write_bad_status(self, unix_timestamp):
        status = dict()
        status['timestamp'] = Reader.to_local_time(unix_timestamp)
        status['error_code'] = 1
        status['error_message'] = 'A dud message'
        status['transaction_count'] = 0
        self.writer.write_status(status)

    def test_getting_status_stdout(self):
        last_good_call_seconds_ago = 300
        last_good_call_minutes = int(round(last_good_call_seconds_ago / 60, 0))
        health = round(100 * (settings.health_list_length - 1) / settings.health_list_length, 1)
        self.write_bad_status(int(time.time()) - 200)
        self.write_good_status(int(time.time()) - last_good_call_seconds_ago)
        output = self.reader.status_request()
        expected_std_status = "Last successful call {} minutes ago, health {}%".format(last_good_call_minutes, health)
        self.assertEqual(expected_std_status, output)

    def test_getting_status_dict(self):
        last_good_call_seconds_ago = 100
        last_good_call_minutes = int(round(last_good_call_seconds_ago / 60, 0))
        health = round(100 * (settings.health_list_length - 1) / settings.health_list_length, 1)
        self.write_bad_status(int(time.time()) - 200)
        self.write_good_status(int(time.time()) - last_good_call_seconds_ago)
        output = self.reader.status_request(as_dict=True)
        expected_status = {'last_call': last_good_call_minutes, 'health': health, 'status': 'Ok'}
        self.assertEqual(expected_status, output)

    def test_getting_status_dict_logger_not_ok(self):
        last_good_call_seconds_ago = 700
        last_good_call_minutes = int(round(last_good_call_seconds_ago / 60, 0))
        health = round(100 * (settings.health_list_length - 1) / settings.health_list_length, 1)
        self.write_bad_status(int(time.time()) - 200)
        self.write_good_status(int(time.time()) - last_good_call_seconds_ago)
        output = self.reader.status_request(as_dict=True)
        expected_status = {'last_call': last_good_call_minutes, 'health': health, 'status': 'Error'}
        self.assertEqual(expected_status, output)

    def test_request_status_with_no_working_directory_returns_none(self):
        reader = Reader()
        expected_status = None
        output = reader.status_request()
        self.assertEqual(expected_status, output)



import logging
import os
import sys
import socket
import subprocess
import time
import pandas as pd
from configchecker import ConfigChecker
from dbops.sqhelper import SQHelper
from whalealert.api.transactions import Transactions
from whalealert.publisher.writer import Writer
from whalealert.publisher.reader import Reader
import whalealert.settings as settings

log = logging.getLogger(__name__)
PROCESS_NAME = 'whaleAlertLogger'


def daemon_already_running():
    processName = PROCESS_NAME
    daemon_already_running._lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        daemon_already_running._lock_socket.bind('\0' + processName)
        log.info("New daemon logging instance started")
        return False
    except Exception as e:
        _ = e
        log.warning("Attempting to start daemon which is already running")
        return True


class WhaleAlert():
    """ Python wrapper for the Whale Watch API"""
    def __init__(self, working_directory=None, log_level=logging.WARNING):
        if working_directory is not None:
            self.__make_directories_as_needed(working_directory)
            self.__setup_logging(working_directory, log_level)
            self.__config = self.__generate_configuration(working_directory)
            self.__database = self.__setup_database(working_directory)
            self.__status = self.__setup_status_file(working_directory)
            self.__writer = Writer(self.__status, self.__database)
            self.__reader = Reader(self.__status, self.__database, self.__config)
            self.__last_data_request_time = int(time.time())
        else:
            self.__config = None
            self.__status = None
            self.__database = None
        log.debug("Started new Whale Alert API wrapper.")
        self.transactions = Transactions()

    def __setup_logging(self, working_directory, log_level):
        logging_file = os.path.join(working_directory, settings.data_file_directory, settings.log_file_name)
        try:
            logging.basicConfig(format='%(asctime)s %(levelname)s %(module)s: %(message)s',
                                datefmt='%m/%d/%Y %I:%M:%S%p',
                                level=log_level,
                                filename=logging_file)
        except Exception as e_r:
            print("Failed to create logging file. Exception '{}'".format(e_r), file=sys.stderr)
            raise

    def __setup_status_file(self, working_directory):
        status = ConfigChecker()
        status.set_expectation(settings.status_file_last_good_call_section_name, settings.status_file_option_timeStamp,
                               str, '')
        status.set_expectation(settings.status_file_last_good_call_section_name,
                               settings.status_file_option_transaction_count, int, 0)

        status.set_expectation(settings.status_file_last_failed_secion_name, settings.status_file_option_timeStamp, str,
                               '')
        status.set_expectation(settings.status_file_last_failed_secion_name, settings.status_file_option_error_code,
                               int, 0)
        status.set_expectation(settings.status_file_last_failed_secion_name, settings.status_file_option_error_message,
                               str, '')

        status.set_expectation(settings.status_file_current_session_section_name,
                               settings.status_file_option_successful_calls, int, 0)
        status.set_expectation(settings.status_file_current_session_section_name,
                               settings.status_file_option_failed_calls, int, 0)
        status.set_expectation(settings.status_file_current_session_section_name,
                               settings.status_file_option_success_rate, float, 0.0)
        status.set_expectation(settings.status_file_current_session_section_name, settings.status_file_option_health,
                               float, 100.0)

        status.set_expectation(settings.status_file_all_time_section_name, settings.status_file_option_successful_calls,
                               int, 0)
        status.set_expectation(settings.status_file_all_time_section_name, settings.status_file_option_failed_calls,
                               int, 0)
        status.set_expectation(settings.status_file_all_time_section_name, settings.status_file_option_success_rate,
                               float, 0.0)

        target_file = os.path.join(working_directory, settings.data_file_directory, settings.status_file_name)
        status.set_configuration_file(target_file)
        status.write_configuration_file(target_file)
        return status

    def __setup_database(self, working_directory):
        file_location = os.path.join(working_directory, settings.data_file_directory, settings.database_file_name)
        database = SQHelper(file_location)
        if not database.exists():
            log.critical("Failed to create required database file, exiting")
            raise
        return database

    def __generate_configuration(self, working_directory):
        self.__make_directories_as_needed(working_directory)

        config = ConfigChecker()
        config.set_expectation(settings.API_section_name, settings.API_option_private_key, str,
                               settings.API_option_privatate_key_default)
        config.set_expectation(settings.API_section_name, settings.API_option_interval, int,
                               settings.API_option_interval_default)
        config.set_expectation(settings.API_section_name, settings.API_option_minimum_value, int,
                               settings.API_option_minimum_value_default)
        config.set_expectation(settings.API_section_name, settings.API_option_historical_limit, int,
                               settings.API_option_historical_limit_default)

        target_directory = os.path.join(working_directory, settings.input_configuation_filename)
        config.set_configuration_file(target_directory)
        config.write_configuration_file(target_directory)
        return config

    def __make_directories_as_needed(self, working_directory):
        target_directory = os.path.join(working_directory, settings.data_file_directory)
        self.__make_dir(target_directory)

    def __make_dir(self, target_directory):
        if not os.path.exists(target_directory):
            try:
                log.debug("Creating new directory '{}'".format(target_directory))
                os.makedirs(target_directory)
            except OSError as e_r:
                log.error("Cannot create directory '{}'. Exception '{}'".format(target_directory, e_r))
                raise

    def get_configuration(self):
        """ Get the configuration used

        Note: This function always returns None if a working_directory is not supplied when the class object is created.

        Returns:
        config (ConfigChecker) if a valid configuration exists.
        None: No configuration exists
        """
        return self.__config

    def get_status(self):
        """ Get the status file used

        Note: This function always returns None if a working_directory is not supplied when the class object is created.

        Returns:
        status (ConfigChecker) if a valid status file exists.
        None: No configuration exists
        """
        return self.__status

    def get_database(self):
        """ Get the database file used

        Note: This function always returns None if a working_directory is not supplied when the class object is created.

        Returns:
        database (SQHelper) if a valid database is connected.
        None: No database is connected.

        """
        return self.__database

    def get_transactions(self, start_time, end_time=None, api_key=None, cursor=None, min_value=500000, limit=100):
        """ Use the Whale Alert API to get the lastest transactions for a given time period

        Parameters:
        start_time (int): A unix time stamp representing the start time to get transactions (exclusive)

        """

        if type(start_time) is not int:
            raise ValueError("Start time must be a unix time stamp integer")
        if end_time is not None and type(end_time) is not int:
            raise ValueError("End time must be a unix time stamp integer")
        if self.__config is None and api_key is None:
            raise ValueError("An API key needs to be supplied to get latest transactions")
        if self.__config is not None and api_key is None:
            api_key = self.__config.get_value(settings.API_section_name, settings.API_option_private_key)
        if self.__config is not None and min_value == 500000:
            min_value = self.__config.get_value(settings.API_section_name, settings.API_option_minimum_value)

        success, transactions, status = self.transactions.get_transactions(start_time, end_time, api_key, cursor,
                                                                           min_value, limit)
        return success, transactions, status

    def write_custom_status(self, status):
        """
        Write a custom status to the status file

        Parameters:
        status (dict) a python dictionary with four keys:
        - timestamp: An iso8601 timestamp
        - error_code: An integer error code. 200 represents success, other values are failures
        - error_message: A description of the error code
        - transaction_count: The number of transaction for this status (Leave at zero if not relevant)

        Returns:
        True: The status was written
        False: An error occured. Errors are sent to the logging module
        """
        if self.__status is None:
            return False
        return self.__writer.write_status(status)

    def fetch_and_store_data(self, start_time, end_time=None, api_key=None, cursor=None, min_value=500000, limit=100):
        if self.__database is None:
            log.error("Trying to fetch data without an API key or working directory (cannot store data).")
            return False
        elif self.__database is not None and api_key is None:
            api_key = self.__config.get_value(settings.API_section_name, settings.API_option_private_key)
            log.debug("Using configuration supplied API key")
        else:
            log.debug("Using overridden API key")

        success, transactions, status = self.transactions.get_transactions(start_time, end_time, api_key, cursor,
                                                                           min_value, limit)
        if success is True and len(transactions) > 0:
            written_ok = self.__writer.write_transactions(transactions)
            if written_ok is False:
                status['error_code'] = 20
                status['error_message'] = "Failed to write transactions to database"
                success = False
        elif success is True and len(transactions) == 0:
            success = False

        self.__writer.write_status(status)
        return success

    def start_daemon(self, force=False, print_output=False):
        if daemon_already_running() and force is False:
            return

        if self.__database is None:
            log.error("Trying to start daemon with no working directory")
            return

        api_key = self.__config.get_value(settings.API_section_name, settings.API_option_private_key)
        historical_limit = self.__config.get_value(settings.API_section_name, settings.API_option_historical_limit)
        request_interval = self.__config.get_value(settings.API_section_name, settings.API_option_interval)

        if request_interval <= 0:
            log.error("Request interval cannot be less than or equal to zero, daemon not starting")
            return
        if historical_limit <= 0:
            log.error("Historical limit cannot be less than or equal to zero, daemon not starting")
            return

        self.__status.set_value(settings.status_file_current_session_section_name,
                                settings.status_file_option_successful_calls, 0)
        self.__status.set_value(settings.status_file_current_session_section_name,
                                settings.status_file_option_failed_calls, 0)
        self.__status.set_value(settings.status_file_current_session_section_name,
                                settings.status_file_option_success_rate, 100.0)
        self.__status.set_value(settings.status_file_current_session_section_name, settings.status_file_option_health,
                                100.0)
        self.__status.write_configuration_file()

        while True:
            if self.transactions.get_last_cursor() is None:
                start_time = self.__find_latest_timestamp()
            else:
                start_time = 0

            end_time = int(time.time())
            if (end_time - start_time) > historical_limit:
                start_time = end_time - historical_limit

            success = self.fetch_and_store_data(start_time,
                                                end_time=end_time,
                                                api_key=api_key,
                                                cursor=self.transactions.get_last_cursor())

            if success is True and print_output:
                print(self.get_new_transaction(pretty=True))

            time.sleep(request_interval)

    def __find_latest_timestamp(self):
        tables = self.__database.get_table_names()
        latest_timestamp = 0
        for table in tables:
            last = self.__database.get_last_time_entry(table)
            try:
                if last['timestamp'] > latest_timestamp:
                    latest_timestamp = last['timestamp']
            except KeyError:
                pass
        if latest_timestamp >= time.time():
            latest_timestamp = time.time()
        return int(latest_timestamp)

    def kill():
        subprocess.Popen(['killall', PROCESS_NAME])

    def to_excel(self, output_file='whaleAlert.xlsx'):
        if self.__database is None:
            return False

        tables = self.__database.get_table_names()
        if len(tables) == 0:
            return False

        df_list = []
        for table in tables:
            df_list.append(self.__database.table_to_df(table))
        writer = pd.ExcelWriter(output_file, engine='openpyxl')
        _ = [A.to_excel(writer, sheet_name="{0}".format(tables[i])) for i, A in enumerate(df_list)]
        writer.save()
        return True

    def data_request(self,
                     start=0,
                     blockchain=None,
                     symbols=None,
                     max_results=20,
                     pretty=False,
                     as_df=False,
                     as_dict=False):

        request = dict(settings.request_format)
        if blockchain is not None:
            request[settings.request_blockchain] = blockchain
        else:
            request[settings.request_blockchain] = ['*']

        if symbols is not None:
            request[settings.request_symbols] = symbols
        else:
            request[settings.request_symbols] = ['*']

        request[settings.request_from_time] = start
        request[settings.request_maximum_results] = max_results
        return self.__reader.data_request(request, pretty, as_df, as_dict)

    def get_new_transaction(self, pretty=False, as_df=False, as_dict=False):
        """Get the transaction returned since the last call to this method

        WARNING: If start_daemon is called with print_output = True, then this method
        will not work as expected

        Parameters:
        pretty (Bool): Use ascii colour codes to format the output.
        as_df (Bool): Return all transactions as a dataframe
        as_dict (Bool): Retun as a {timestamp: '', 'text' ''} dictionary. Pretty output is also applied

        as_df takes precendence over as_dict.

        Returns:
        Formatted output depending on the passed parameters.
        """
        if as_df is True:
            return self.__writer.get_last_written()

        return self.__reader.dataframe_to_transaction_output(self.__writer.get_last_written(),
                                                             pretty=pretty,
                                                             as_dict=as_dict)

    def status_request(self, as_dict=False):
        return self.__reader.status_request(as_dict=as_dict)



import unittest
from unittest import mock
import json
import time
import logging
from requests.models import Response
from requests import Session
from requests.exceptions import Timeout, TooManyRedirects
import whalealert.settings as settings
from whalealert.api.transactions import Transactions

logging.disable(logging.CRITICAL)

text_error = {"result": "error", "message": "invalid api_key"}
text_error_bad_key = {"result": "error", "bad_key": "invalid api_key"}
text_empty = {"result": "success", "cursor": "2712e8b6-2712e8b6-5eafc647", "count": 0}
text_empty_bad_key = {"result": "success", "bad_key": "2712e8b6-2712e8b6-5eafc647", "count": 0}
text_success = {
    "result":
    "success",
    "cursor":
    "2712f286-2712f286-5eafc711",
    "count":
    2,
    "transactions": [{
        "blockchain": "ethereum",
        "symbol": "usdt",
        "id": "655552612",
        "transaction_type": "transfer",
        "hash": "4cdfc57c737b4214fbce384e57f50d8b52f80c2eb470654f9fdc2062c534bf42",
        "from": {
            "address": "477b8d5ef7c2c42db84deb555419cd817c336b6f",
            "owner_type": "unknown"
        },
        "to": {
            "address": "b3fe1649862d7889ab002e0224a2db54870eafa9",
            "owner_type": "unknown"
        },
        "timestamp": 1588578025,
        "amount": 500000,
        "amount_usd": 503430.84,
        "transaction_count": 1
    }, {
        "blockchain": "ethereum",
        "symbol": "USDT",
        "id": "655553158",
        "transaction_type": "transfer",
        "hash": "19a393cb0fe2cd5975a3741e43f05859877b639b1c93dd19acd4abfa08715530",
        "from": {
            "address": "df38a19aa5db1f15c6df389d86175285d45fa572",
            "owner_type": "exchange",
            "owner": "stuart"
        },
        "to": {
            "address": "80c6e081ae5813b163357774003d3695faa0f53e",
            "owner_type": "exchange",
            "owner": "stuart"
        },
        "timestamp": 1588578065,
        "amount": 500000,
        "amount_usd": 509513.7,
        "transaction_count": 1
    }]
}
text_success_bad_count = {
    "result":
    "success",
    "cursor":
    "2712f286-2712f286-5eafc711",
    "count":
    3,
    "transactions": [{
        "blockchain": "ethereum",
        "symbol": "usdt",
        "id": "655552612",
        "transaction_type": "transfer",
        "hash": "4cdfc57c737b4214fbce384e57f50d8b52f80c2eb470654f9fdc2062c534bf42",
        "from": {
            "address": "477b8d5ef7c2c42db84deb555419cd817c336b6f",
            "owner_type": "unknown"
        },
        "to": {
            "address": "b3fe1649862d7889ab002e0224a2db54870eafa9",
            "owner_type": "unknown"
        },
        "timestamp": 1588578025,
        "amount": 500000,
        "amount_usd": 503430.84,
        "transaction_count": 1
    }, {
        "blockchain": "ethereum",
        "symbol": "usdt",
        "id": "655553158",
        "transaction_type": "transfer",
        "hash": "19a393cb0fe2cd5975a3741e43f05859877b639b1c93dd19acd4abfa08715530",
        "from": {
            "address": "df38a19aa5db1f15c6df389d86175285d45fa572",
            "owner_type": "unknown"
        },
        "to": {
            "address": "80c6e081ae5813b163357774003d3695faa0f53e",
            "owner_type": "unknown"
        },
        "timestamp": 1588578065,
        "amount": 500000,
        "amount_usd": 509513.7,
        "transaction_count": 1
    }]
}
text_success_bad_key = {
    "result":
    "success",
    "cursor":
    "2712f286-2712f286-5eafc711",
    "count":
    2,
    "transactions": [{
        "bad_key": "ethereum",
        "symbol": "usdt",
        "id": "655552612",
        "transaction_type": "transfer",
        "hash": "4cdfc57c737b4214fbce384e57f50d8b52f80c2eb470654f9fdc2062c534bf42",
        "from": {
            "address": "477b8d5ef7c2c42db84deb555419cd817c336b6f",
            "owner_type": "unknown"
        },
        "to": {
            "address": "b3fe1649862d7889ab002e0224a2db54870eafa9",
            "owner_type": "unknown"
        },
        "timestamp": 1588578025,
        "amount": 500000,
        "amount_usd": 503430.84,
        "transaction_count": 1
    }, {
        "blockchain": "ethereum",
        "symbol": "usdt",
        "id": "655553158",
        "transaction_type": "transfer",
        "hash": "19a393cb0fe2cd5975a3741e43f05859877b639b1c93dd19acd4abfa08715530",
        "from": {
            "address": "df38a19aa5db1f15c6df389d86175285d45fa572",
            "owner_type": "unknown"
        },
        "to": {
            "address": "80c6e081ae5813b163357774003d3695faa0f53e",
            "owner_type": "unknown"
        },
        "timestamp": 1588578065,
        "amount": 500000,
        "amount_usd": 509513.7,
        "transaction_count": 1
    }]
}
text_bad_json = 'not a good json format'


class WhaleAlertAPI(unittest.TestCase):
    def setUp(self):
        self.transactions = Transactions()

    def test_correctly_formed_api_call_minimum_parameters(self):
        r = Response
        r.status_code = 200
        r.text = text_empty
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)

        api_key = '123'
        start_time = 123456
        min_value = 100000
        end_time = None
        cursor = None
        limit = 100
        call_parameters = {}
        call_parameters['api_key'] = api_key
        call_parameters['start'] = start_time
        call_parameters['min_value'] = min_value
        call_parameters['limit'] = 100

        expected = [
            mock.call(settings.whale_get_transactions_url,
                      params=call_parameters,
                      timeout=settings.whale_call_timeout_seconds)
        ]
        self.transactions.get_transactions(start_time, end_time, api_key, cursor, min_value, limit)
        self.assertEqual(expected, Session.get.mock_calls)

    def test_correctly_formed_api_call_all_parameters(self):
        r = Response
        r.status_code = 200
        r.text = text_empty
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)

        api_key = '123'
        start_time = 123456
        min_value = 100000
        end_time = 234567
        cursor = 'asdf'
        limit = 100
        call_parameters = {}
        call_parameters['api_key'] = api_key
        call_parameters['start'] = start_time
        call_parameters['min_value'] = min_value
        call_parameters['limit'] = 100
        call_parameters['end'] = end_time
        call_parameters['cursor'] = cursor

        expected = [
            mock.call(settings.whale_get_transactions_url,
                      params=call_parameters,
                      timeout=settings.whale_call_timeout_seconds)
        ]
        self.transactions.get_transactions(start_time, end_time, api_key, cursor, min_value, limit)
        self.assertEqual(expected, Session.get.mock_calls)

    def test_handle_connection_error_exception(self):
        Session.get = mock.MagicMock().method()
        Session.get.side_effect = ConnectionError
        time.sleep = mock.MagicMock()
        success, transactions, status = self.make_nomral_request()
        call_count = len(Session.get.call_args_list)
        self.assertEqual(call_count, settings.whale_retries_on_failure + 1)
        self.assertEqual(success, False)

    def test_handle_timeouterror_exception(self):
        Session.get = mock.MagicMock().method()
        Session.get.side_effect = Timeout
        time.sleep = mock.MagicMock()
        success, transactions, status = self.make_nomral_request()
        call_count = len(Session.get.call_args_list)
        self.assertEqual(call_count, settings.whale_retries_on_failure + 1)
        self.assertEqual(success, False)

    def test_handle_too_many_redirects_exception(self):
        Session.get = mock.MagicMock().method()
        Session.get.side_effect = TooManyRedirects
        time.sleep = mock.MagicMock()
        success, transactions, status = self.make_nomral_request()
        call_count = len(Session.get.call_args_list)
        self.assertEqual(call_count, settings.whale_retries_on_failure + 1)
        self.assertEqual(success, False)

    def test_handle_other_exception(self):
        Session.get = mock.MagicMock().method()
        Session.get.side_effect = ValueError
        time.sleep = mock.MagicMock()
        success, transactions, status = self.make_nomral_request()
        call_count = len(Session.get.call_args_list)
        self.assertEqual(call_count, settings.whale_retries_on_failure + 1)
        self.assertEqual(success, False)

    def test_bad_json_response_handled_ok(self):
        r = Response
        r.status_code = 200
        r.text = text_bad_json
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 5)

    def test_parsing_error_response(self):
        r = Response
        r.status_code = 401
        r.text = json.dumps(text_error)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 401)

    def test_parsing_error_response_bad_key(self):
        r = Response
        r.status_code = 401
        r.text = json.dumps(text_error_bad_key)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 7)

    def test_parsing_good_response_no_transactions(self):
        r = Response
        r.status_code = 200
        r.text = json.dumps(text_empty)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, True)
        self.assertEqual(status[settings.status_file_option_error_code], 200)

    def test_parsing_good_response_no_transactions_bad_key(self):
        r = Response
        r.status_code = 200
        r.text = json.dumps(text_empty_bad_key)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 9)

    def test_parsing_good_response_with_transactions_wrong_count(self):
        r = Response
        r.status_code = 200
        r.text = json.dumps(text_success_bad_count)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 8)

    def test_parsing_good_response_with_transactions(self):
        r = Response
        r.status_code = 200
        r.text = json.dumps(text_success)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, True)
        self.assertEqual(status[settings.status_file_option_error_code], 200)
        self.assertGreaterEqual(self.transactions.get_last_timestamp(), int(time.time()) - 1)
        self.assertEqual(self.transactions.get_last_cursor(), text_success['cursor'])
        self.assertEqual(transactions[1], text_success['transactions'][1])

    def test_parsing_good_response_with_transactions_bad_key(self):
        r = Response
        r.status_code = 200
        r.text = json.dumps(text_success_bad_key)
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 10)

    def test_bad_unknown_json_parse_error_handled(self):
        r = Response
        r.status_code = 200
        r.text = text_bad_json
        Session.get = mock.MagicMock().method()
        Session.get.return_value = (r)
        old_jsonload = json.loads
        json.loads = mock.MagicMock().method()
        json.loads.side_effect = ValueError
        success, transactions, status = self.make_nomral_request()
        self.assertEqual(success, False)
        self.assertEqual(status[settings.status_file_option_error_code], 6)
        json.loads = old_jsonload

    def make_nomral_request(self):
        api_key = '123'
        start_time = 123456
        min_value = 500000
        end_time = 234567
        cursor = 'asdf'
        limit = 100
        return self.transactions.get_transactions(start_time, end_time, api_key, cursor, min_value, limit)



import logging
import pandas as pd
from configchecker import ConfigChecker
import whalealert.settings as settings

log = logging.getLogger(__name__)


class Writer():
    """ Puslishing Whale Alert API status and database results"""
    def __init__(self, status, database):
        self.__status = status
        self.__database = database
        self.__health_list = [1] * settings.health_list_length
        self.__last_written = []
        if status is not None:
            log.debug("Pulisher started with initial status {}".format(status.get_expectations()))

    def get_status_object(self):
        """ Return the status object used"""
        return self.__status

    def get_database(self):
        """Return the database object used"""
        return self.__database

    def write_transactions(self, transactions):
        """
        Write transactions to the dataabse

        Parameters:
        transactions (dict): Transactions returned by api.get_trasactions

        Returns:
        True: Transactions were written successfully
        False: An error occured, status written to logs.
        """
        if len(transactions) == 0:
            log.debug("Trying to write an empty list of transactions")
            return False

        for transaction in transactions:
            try:
                transaction = self.__squash_dictionary(transaction)
                self.__create_tables_as_needed(transaction)
            except KeyError:
                log.error("Key error parsing keys from transaction {}".format(transaction))
                return False
            except Exception as e:
                log.error("Exception {} when parsing keys from transaction {}".format(e, transaction))
                return False
            if self.__add_entries(transaction) is False:
                log.error("Failed to add entriy to database. Entry {}".format(transaction))
                return False
        return True

    def get_last_written(self):
        current_transactions = list(self.__last_written)
        self.__last_written.clear()
        return pd.DataFrame(current_transactions)

    def __add_entries(self, transaction):
        table_name = transaction[settings.database_table_identifier]
        if len(self.__last_written) < settings.maximum_stored_latest_transaction:
            self.__last_written.append(transaction)
        return self.__database.insert(table_name, transaction)

    def __squash_dictionary(self, transaction):
        new_dict = dict(transaction)
        new_dict[settings.database_column_from_address] = transaction['from']['address']
        new_dict[settings.database_column_from_owner] = transaction['from']['owner']
        new_dict[settings.database_column_from_owner_type] = transaction['from']['owner_type']
        new_dict[settings.database_column_to_address] = transaction['to']['address']
        new_dict[settings.database_column_to_owner] = transaction['to']['owner']
        new_dict[settings.database_column_to_owner_type] = transaction['to']['owner_type']

        new_dict.pop('from', None)
        new_dict.pop('to', None)
        return new_dict

    def __create_tables_as_needed(self, transaction):
        self.__database.create_table(transaction[settings.database_table_identifier], settings.database_columns)

    def write_status(self, status):
        """"
        Write the logger status to the status file

        Returns:
        True: The file was written successfully
        False: The file couldn't be written, error status written to logs
        """
        if type(self.__status) is not ConfigChecker:
            return False
        try:
            self.__update_all_time(status)
            self.__update_current(status)
            self.__update_good_call_stats(status)
            self.__update_bad_call_stats(status)
        except KeyError:
            log.error("Key error when trying to write status {}".format(status))
            return False

        if not self.__status.write_configuration_file():
            log.error("Failed to write status file")
            return False
        return True

    def __update_good_call_stats(self, status):
        if status[settings.status_file_option_error_code] != 200:
            return
        self.__status.set_value(settings.status_file_last_good_call_section_name, settings.status_file_option_timeStamp,
                                status[settings.status_file_option_timeStamp])
        self.__status.set_value(settings.status_file_last_good_call_section_name,
                                settings.status_file_option_transaction_count,
                                status[settings.status_file_option_transaction_count])

    def __update_bad_call_stats(self, status):
        if status[settings.status_file_option_error_code] == 200:
            return
        self.__status.set_value(settings.status_file_last_failed_secion_name, settings.status_file_option_timeStamp,
                                status[settings.status_file_option_timeStamp])
        self.__status.set_value(settings.status_file_last_failed_secion_name, settings.status_file_option_error_code,
                                status[settings.status_file_option_error_code])
        self.__status.set_value(settings.status_file_last_failed_secion_name, settings.status_file_option_error_message,
                                status[settings.status_file_option_error_message])

    def __update_all_time(self, status):
        if status[settings.status_file_option_error_code] == 200:
            self.__increment_value(settings.status_file_all_time_section_name,
                                   settings.status_file_option_successful_calls)
        else:
            self.__increment_value(settings.status_file_all_time_section_name, settings.status_file_option_failed_calls)

        self.__calculate_success_rate(settings.status_file_all_time_section_name)

    def __update_current(self, status):
        if status[settings.status_file_option_error_code] == 200:
            self.__increment_value(settings.status_file_current_session_section_name,
                                   settings.status_file_option_successful_calls)
            self.__calculate_health(True)
        else:
            self.__increment_value(settings.status_file_current_session_section_name,
                                   settings.status_file_option_failed_calls)
            self.__calculate_health(False)
        self.__calculate_success_rate(settings.status_file_current_session_section_name)
        return True

    def __calculate_success_rate(self, section):
        good_calls = self.__status.get_value(section, settings.status_file_option_successful_calls)
        bad_calls = self.__status.get_value(section, settings.status_file_option_failed_calls)
        percent = 100 * good_calls / (good_calls + bad_calls)
        self.__status.set_value(section, settings.status_file_option_success_rate, round(percent, 2))

    def __calculate_health(self, success):
        self.__health_list.append((int(success)))
        self.__health_list.pop(0)
        health = round((100 * sum(self.__health_list) / len(self.__health_list)), 1)
        self.__status.set_value(settings.status_file_current_session_section_name, settings.status_file_option_health,
                                health)

    def __increment_value(self, section, key):
        self.__status.set_value(section, key, self.__status.get_value(section, key) + 1)



# Filenames
input_configuation_filename = "config.ini"
data_file_directory = "data"
log_file_directory = data_file_directory
status_file_directory = data_file_directory
log_file_name = "log"
database_file_name = 'whaleAlert.db'
status_file_name = 'status.ini'
appNameDirectory = 'whaleAlertLogger'

# Input configuration file
API_section_name = "API"
API_option_private_key = "api_private_key"
API_option_privatate_key_default = 'your-private-key-here'
API_option_interval = 'request_interval_seconds'
API_option_interval_default = 60
API_option_minimum_value = 'minimum_transaction_value'
API_option_minimum_value_default = 500000
API_option_historical_limit = 'historical_limit'
API_option_historical_limit_default = 3599

# Status file definitions
status_file_last_good_call_section_name = 'Last Successful Call'
status_file_last_failed_secion_name = 'Last Failed Call'

status_file_option_transaction_count = 'transaction_count'
status_file_option_timeStamp = 'timestamp'
status_file_option_error_code = 'error_code'
status_file_option_error_message = 'error_message'

status_file_all_time_section_name = 'All Time'
status_file_current_session_section_name = 'Current Session'

status_file_option_successful_calls = 'successful_calls'
status_file_option_failed_calls = 'failed_calls'
status_file_option_success_rate = 'success_rate'
status_file_option_health = "health"

# Database definitions
database_column_blockchain = 'blockchain'
database_column_symbol = 'symbol'
database_column_id = 'id'
database_column_transaction_type = 'transaction_type'
database_column_hash = 'hash'
database_column_from_address = 'from_address'
database_column_from_owner = 'from_owner'
database_column_from_owner_type = 'from_owner_type'
database_column_to_address = 'to_address'
database_column_to_owner = 'to_owner'
database_column_to_owner_type = 'to_owner_type'
database_column_timestamp = 'timestamp'
database_column_amount = 'amount'
database_column_amount_usd = 'amount_usd'
database_column_transaction_count = 'transaction_count'

database_table_identifier = 'blockchain'

database_columns = {
    database_column_blockchain: 'TEXT',
    database_column_symbol: 'TEXT',
    database_column_id: 'TEXT',
    database_column_transaction_type: 'TEXT',
    database_column_hash: 'TEXT',
    database_column_from_address: 'TEXT',
    database_column_from_owner: 'TEXT',
    database_column_from_owner_type: 'TEXT',
    database_column_to_address: 'TEXT',
    database_column_to_owner: 'TEXT',
    database_column_to_owner_type: 'TEXT',
    database_column_timestamp: 'NUMERIC',
    database_column_amount: 'REAL',
    database_column_amount_usd: 'REAL',
    database_column_transaction_count: 'NUMERIC'
}

# Whale Alert API
whale_get_transactions_url = 'https://api.whale-alert.io/v1/transactions'
whale_retries_on_failure = 3
whale_call_timeout_seconds = 10

whale_error_message = 'message'

whale_success_result = 'result'
whale_success_cursor = 'cursor'
whale_success_count = 'count'
whale_success_transactions = 'transactions'

whale_transaction_blockchain = 'blockchain'
whale_transaction_symbol = 'symbol'
whale_transaction_id = 'id'
whale_transaction_transaction_type = 'transaction_type'
whale_transaction_hash = 'hash'
whale_transaction_from = 'from'
whale_transaction_address = 'address'
whale_transaction_owner = 'owner'
whale_transaction_owner_type = 'owner_type'
whale_transaction_to = 'to'
whale_transaction_timestamp = 'timestamp'
whale_transaction_amount = 'amount'
whale_transaction_amount_usd = 'amount_usd'
whale_transaction_transaction_count = 'transaction_count'

# Puslisher settings
health_list_length = 30

# Data request settings
request_blockchain = 'blockchain'
request_symbols = 'symbols'
request_from_time = 'from_time'
request_output_format = 'output_format'
request_maximum_results = 'maximum_results'

request_format = {
    request_blockchain: [],
    request_symbols: [],
    request_from_time: 0,
    request_output_format: 'str',
    request_maximum_results: 1
}

request_time_format = "%m/%d/%Y %H:%M:%S"

maximum_stored_latest_transaction = 1000



import logging
import time
import datetime
import pandas as pd
from colorama import Fore
from colorama import Style
import whalealert.settings as settings
from dateutil import tz, parser

log = logging.getLogger(__name__)
UNDERLINE = '\033[4m'


class Reader():
    """ Reading Whale Alert API status and database results"""
    def __init__(self, status=None, database=None, config=None):
        self.__status = status
        self.__database = database
        self.__config = config
        if status is not None:
            log.debug("Pulisher started with initial status {}".format(status.get_expectations()))

    def get_status_object(self):
        """ Return the status object used"""
        return self.__status

    def get_database(self):
        """Return the database object used"""
        return self.__database

    def status_request(self, as_dict=False):
        if self.__status is None or self.__config is None:
            return None

        minutes = self.__calculate_mintues_since_last_good_call()
        health = self.__status.get_value(settings.status_file_current_session_section_name,
                                         settings.status_file_option_health)
        if as_dict:
            status = dict()
            if self.__logger_status_ok():
                status['status'] = "Ok"
            else:
                status['status'] = "Error"
            status['health'] = health
            status['last_call'] = minutes
            return status
        else:
            return "Last successful call {} minutes ago, health {}%".format(minutes, health)

    def __logger_status_ok(self):
        call_inteval = self.__config.get_value(settings.API_section_name, settings.API_option_interval)
        last_call = Reader.from_local_time(
            self.__status.get_value(settings.status_file_last_good_call_section_name,
                                    settings.status_file_option_timeStamp))
        if (int(time.time()) - last_call) > (call_inteval * 5):
            return False
        return True

    def __calculate_mintues_since_last_good_call(self):
        lastCall = Reader.from_local_time(
            self.__status.get_value(settings.status_file_last_good_call_section_name,
                                    settings.status_file_option_timeStamp))
        currentTime = int(time.time())
        return int(round((currentTime - lastCall) / 60, 0))

    def data_request(self, request, pretty=False, as_df=False, as_dict=False):

        if self.get_database() is None:
            log.error("Requesting data from a database with no working directory")
            return self.__return_empty_result(pretty, as_df, as_dict)

        if self.__check_data_request_keys(request) is False:
            return self.__return_empty_result(pretty, as_df, as_dict)

        entries = self.__filter_request_by_blockchain(request)
        entries = self.__filter_request_by_symbols(request, entries)

        if len(entries) == 0:
            return self.__return_empty_result(pretty, as_df, as_dict)

        return self.dataframe_to_transaction_output(entries, request, pretty, as_dict=as_dict, as_df=as_df)

    def __filter_request_by_symbols(self, request, entries):
        symbols = request[settings.request_symbols]
        if symbols == ['*']:
            filter_by_symbol = entries
        else:
            filter_by_symbol = entries[entries[settings.database_column_symbol].isin(request[settings.request_symbols])]
        return filter_by_symbol

    def __filter_request_by_blockchain(self, request):
        all_entries = pd.DataFrame()
        if request[settings.request_blockchain] == ['*']:
            blockchains = self.__database.get_table_names()
        else:
            blockchains = request[settings.request_blockchain]

        for blockchain in blockchains:
            new_entries = self.__database.get_row_range(blockchain, 'timestamp',
                                                        request[settings.request_from_time] + 1, int(time.time()))
            if new_entries is None:
                log.warning("Data request for blockchain {} which isn't in database".format(blockchain))
                continue
            all_entries = all_entries.append(new_entries)
        return all_entries

    def dataframe_to_transaction_output(self, df, request=None, pretty=False, as_dict=False, as_df=False):

        sorted_by_time = df.sort_values('timestamp', ascending=True)

        if request is not None and request[settings.request_maximum_results] >= 0:
            sorted_by_time = sorted_by_time.tail(request[settings.request_maximum_results])

        if as_df:
            sorted_by_time.reset_index(drop=True, inplace=True)
            return sorted_by_time
        return self.__make_result_string(sorted_by_time, pretty, as_dict)

    def __return_empty_result(self, pretty, as_df, as_dict):
        if as_df:
            return pd.DataFrame()
        elif as_dict:
            return []
        else:
            return ''

    def __make_result_string(self, results, pretty, as_dict):

        output = ''
        output_list = []
        for index, result in results.iterrows():
            try:
                if result[settings.database_column_to_owner] == '':
                    result[settings.database_column_to_owner] = 'unknown'
                if result[settings.database_column_from_owner] == '':
                    result[settings.database_column_from_owner] = 'unknown'
                if as_dict:
                    output_dict = dict()
                    output_dict['timestamp'] = self.__make_time_string(result, pretty)[:-1]
                    output_dict['text'] = self.__make_amount_string(result, pretty) + self.__make_transfer_string(
                        result, pretty)[:-1]
                    output_list.append(output_dict)
                else:
                    output = output + self.__make_time_string(result, pretty) + self.__make_amount_string(
                        result, pretty) + self.__make_transfer_string(result, pretty)
            except Exception as e:
                log.error("Invalid column names found in database. Exception {}".format(e))

        log.debug("Successful data request returned {} results".format(len(results)))
        if as_dict:
            return output_list
        else:
            return output[:-1]

    def __make_time_string(self, result, pretty):
        if pretty:
            return Fore.YELLOW + Reader.to_local_time(
                result[settings.database_column_timestamp]) + " " + Style.RESET_ALL
        else:
            return Reader.to_local_time(result[settings.database_column_timestamp]) + " "

    def __make_amount_string(self, result, pretty):
        if pretty:
            output = Fore.WHITE + str("{:.2f}".format(
                result[settings.database_column_amount])) + " " + Fore.RED + result[settings.database_column_symbol]
            if result[settings.database_column_amount_usd] < 1000000:
                output = output + Fore.CYAN + " (" + Reader.format_currency(
                    result[settings.database_column_amount_usd]) + " USD) "
            elif result[settings.database_column_amount_usd] < 20000000:
                output = output + Fore.CYAN + Style.BRIGHT + " (" + Reader.format_currency(
                    result[settings.database_column_amount_usd]) + " USD) "
            else:
                output = output + ' ' + Fore.CYAN + Style.BRIGHT + UNDERLINE + "(" + Reader.format_currency(
                    result[settings.database_column_amount_usd]) + " USD)"
            return output + Style.RESET_ALL + ' '
        else:
            return str("{:.2f}".format(result[settings.database_column_amount])) + " " + result[
                settings.database_column_symbol] + " (" + Reader.format_currency(
                    result[settings.database_column_amount_usd]) + " USD) "

    def __make_transfer_string(self, result, pretty):
        if pretty:
            if result[settings.database_column_transaction_type] == 'transfer':
                output = 'transferred from ' + Fore.BLUE + result[settings.database_column_from_owner]  \
                        + Style.RESET_ALL + ' to ' + Fore.BLUE + result[settings.database_column_to_owner] \
                         + Style.RESET_ALL + '.\n'
            elif result[settings.database_column_transaction_type] == 'burn':
                output = Fore.RED + 'burned' + Style.RESET_ALL + ' at ' +  \
                        Fore.BLUE + result[settings.database_column_from_owner] + Style.RESET_ALL + '.\n'
            else:
                output = result[settings.database_column_transaction_type] + ' from ' +  \
                        Fore.BLUE + result[settings.database_column_from_owner]  \
                        + Style.RESET_ALL + ' to ' + Fore.BLUE + result[settings.database_column_to_owner] \
                        + Style.RESET_ALL + '.\n'

            return output
        else:
            return "transferred from " + result[settings.database_column_from_owner] + \
                    ' to ' + result[settings.database_column_to_owner] + '.' + '\n'

    def __check_data_request_keys(self, request):
        try:
            dummy = request[settings.request_blockchain]
            dummy = request[settings.request_from_time]
            dummy = request[settings.request_symbols]
            dummy = request[settings.request_maximum_results]
            dummy = request[settings.request_output_format]
            _ = dummy
        except KeyError:
            log.error("Invalid keys supplied in data request {}".format(request))
            return False
        return True

    def format_currency(value):
        return str(f'{value:,.2f}'.replace('$-', '-$'))

    def to_local_time(unix_timestamp):
        return datetime.datetime.fromtimestamp(unix_timestamp).astimezone(tz.tzlocal()).strftime(
            settings.request_time_format)

    def from_local_time(iso8601):
        return int((parser.parse(iso8601).astimezone(tz.tzlocal()).strftime('%s')))



import setuptools
exec(open('whalealert/_version.py').read())

with open("README.md","r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "whale-alert",
    scripts = ['bin/whaleAlertLogger'],
    version = '0.0.1',
    author ="stuianna",
    author_email = "stuian@protonmail.com",
    description = "Cryptocurrency Whale Alert API Logger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = "https://github.com/stuianna/whaleAlert",
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
        ],
    python_requires = '>=3.6',
    install_requires=[
        'requests',
        'colorama',
        'config-checker',
        'db-ops',
        'requests',
        'pandas',
        'numpy',
        'appdirs',
        'python-dateutil',
        'urllib3'
    ],
    )



import unittest
from unittest import mock
import logging
import os
from requests.models import Response
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import whalealert.settings as settings
from whalealert.whalealert import WhaleAlert
from whalealert.api.transactions import Transactions
import json
import datetime

TEST_WORKING_DIR = os.path.join('tests', 'test_working_directory')
TEST_WORKING_DIR_BAD_PERMISSIONS = "/root/test"
TEST_WORKING_DIR_ALREADY_EXISTS = "tests/test_whale_alert.py"

good_status = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 200,
    'error_message': '',
    'transaction_count': 3
}

bad_status = {
    'timestamp': '2020-05-04T16:20:37.276101',
    'error_code': 401,
    'error_message': 'Bad API Key',
    'transaction_count': 0
}

bad_formatted_status = {}

good_transactions = [{
    'blockchain': 'bitcoin',
    'symbol': 'btc',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1,
}]

bad_transactions = [{
    'blockchain': 'bitcoin',
    'symbol': 'btc',
    'id': '662472177',
    'transaction_type': 'transfer',
    'hash': '8d5ae34805f70d0a412964dca4dbd3f48bc103700686035a61b293cb91fe750d',
    'from': {
        'address': 'f2103b01cd7957f3a9d9726bbb74c0ccd3f355d3',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'to': {
        'address': '3f5ce5fbfe3e9af3971dd833d26ba9b5c936f0be',
        'owner': 'binance',
        'owner_type': 'exchange'
    },
    'timestamp': 1588874414,
    'amount': 3486673,
    'amount_usd': 3508660.2,
    'transaction_count': 1,
    'bad_key': 1,
}]


def cleanup_working_directories():
    try:
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.database_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.data_file_directory, settings.status_file_name))
        os.remove(os.path.join(TEST_WORKING_DIR, settings.input_configuation_filename))
        os.removedirs(os.path.join(TEST_WORKING_DIR, settings.data_file_directory))
        os.removedirs(TEST_WORKING_DIR)
    except Exception as e_r:
        _ = e_r


class WhaleAlertInit(unittest.TestCase):
    """ Testing whale alert initialisation"""
    def setUp(self):
        self.whale = WhaleAlert(TEST_WORKING_DIR, log_level=logging.DEBUG)
        self.config = self.whale.get_configuration()
        self.status = self.whale.get_status()
        self.database = self.whale.get_database()

    def tearDown(self):
        cleanup_working_directories()

    def test_configuration_directories_and_files_created_ok_with_good_directory(self):
        whale = WhaleAlert(working_directory=TEST_WORKING_DIR, log_level=logging.DEBUG)
        self.check_directories_are(TEST_WORKING_DIR, True)

    def test_configuration_directories_and_files_created_ok_with_bad_directory_permissions(self):
        self.assertRaises(PermissionError, WhaleAlert, TEST_WORKING_DIR_BAD_PERMISSIONS, log_level=logging.DEBUG)

    def test_configuration_directories_and_files_created_ok_with_bad_directory_permissions(self):
        self.assertRaises(NotADirectoryError, WhaleAlert, TEST_WORKING_DIR_ALREADY_EXISTS, log_level=logging.DEBUG)

    def check_directories_are(self, directory, status):
        config_exists = os.path.exists(os.path.join(directory, settings.input_configuation_filename))
        data_dir_exists = os.path.exists(os.path.join(directory, settings.data_file_directory))
        database_exists = os.path.exists(
            os.path.join(directory, settings.data_file_directory, settings.database_file_name))
        status_exists = os.path.exists(os.path.join(directory, settings.data_file_directory, settings.status_file_name))
        logs_exists = os.path.exists(os.path.join(directory, settings.data_file_directory, settings.log_file_name))
        self.assertEqual(config_exists, status)
        self.assertEqual(data_dir_exists, status)
        self.assertEqual(database_exists, status)
        self.assertEqual(status_exists, status)

    def test_status_database_config_are_none_with_no_working_directory(self):
        whale = WhaleAlert()
        config = whale.get_configuration()
        status = whale.get_status()
        database = whale.get_database()
        self.assertIs(config, None)
        self.assertIs(status, None)
        self.assertIs(database, None)

    def test_config_values_are_default(self):
        self.assertEqual(self.config.get_value(settings.API_section_name, settings.API_option_private_key),
                         settings.API_option_privatate_key_default)
        self.assertEqual(self.config.get_value(settings.API_section_name, settings.API_option_interval),
                         settings.API_option_interval_default)
        self.assertEqual(self.config.get_value(settings.API_section_name, settings.API_option_minimum_value),
                         settings.API_option_minimum_value_default)
        self.assertEqual(self.config.get_value(settings.API_section_name, settings.API_option_historical_limit),
                         settings.API_option_historical_limit_default)

    def test_status_values_are_default(self):
        self.assertEqual(
            self.status.get_value(settings.status_file_last_good_call_section_name,
                                  settings.status_file_option_timeStamp), '')
        self.assertEqual(
            self.status.get_value(settings.status_file_last_good_call_section_name,
                                  settings.status_file_option_transaction_count), 0)

        self.assertEqual(
            self.status.get_value(settings.status_file_last_failed_secion_name, settings.status_file_option_timeStamp),
            '')
        self.assertEqual(
            self.status.get_value(settings.status_file_last_failed_secion_name, settings.status_file_option_error_code),
            0)
        self.assertEqual(
            self.status.get_value(settings.status_file_last_failed_secion_name,
                                  settings.status_file_option_error_message), '')

        self.assertEqual(
            self.status.get_value(settings.status_file_current_session_section_name,
                                  settings.status_file_option_successful_calls), 0)
        self.assertEqual(
            self.status.get_value(settings.status_file_current_session_section_name,
                                  settings.status_file_option_failed_calls), 0)
        self.assertEqual(
            self.status.get_value(settings.status_file_current_session_section_name,
                                  settings.status_file_option_success_rate), 0.0)
        self.assertEqual(
            self.status.get_value(settings.status_file_current_session_section_name,
                                  settings.status_file_option_health), 100.0)

        self.assertEqual(
            self.status.get_value(settings.status_file_all_time_section_name,
                                  settings.status_file_option_successful_calls), 0)
        self.assertEqual(
            self.status.get_value(settings.status_file_all_time_section_name, settings.status_file_option_failed_calls),
            0)
        self.assertEqual(
            self.status.get_value(settings.status_file_all_time_section_name, settings.status_file_option_success_rate),
            0.0)


class WhaleAlertAPICall(unittest.TestCase):
    """ Testing setup for whale alert api call"""
    def setUp(self):
        self.whale = WhaleAlert(TEST_WORKING_DIR)
        self.config = self.whale.get_configuration()
        self.status = self.whale.get_status()
        self.database = self.whale.get_database()

    def tearDown(self):
        cleanup_working_directories()

    def test_configuration_values_are_used_if_working_directory_supplied(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (False, {}, {})
        self.config.set_value(settings.API_section_name, settings.API_option_private_key, '1234')
        self.config.set_value(settings.API_section_name, settings.API_option_minimum_value, 10)
        expected = [mock.call(0, None, '1234', None, 10, 100)]
        self.whale.get_transactions(0)
        self.assertEqual(self.whale.transactions.get_transactions.mock_calls, expected)

    def test_configuration_values_are_overwrite_config_values_if_supplied(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (False, {}, {})
        self.config.set_value(settings.API_section_name, settings.API_option_private_key, '1234')
        self.config.set_value(settings.API_section_name, settings.API_option_minimum_value, 10)
        expected = [mock.call(0, None, '7890', None, 500, 100)]
        self.whale.get_transactions(0, api_key='7890', min_value=500)
        self.assertEqual(self.whale.transactions.get_transactions.mock_calls, expected)

    def test_no_config_value_raises_exception_if_api_key_not_supplied(self):
        whale = WhaleAlert()
        self.assertRaises(ValueError, whale.get_transactions, 0)

    def test_raises_exception_if_start_time_is_not_int(self):
        whale = WhaleAlert()
        self.assertRaises(ValueError, whale.get_transactions, 'not_an_int', api_key='1234')

    def test_raises_exception_if_end_time_is_not_int(self):
        whale = WhaleAlert()
        self.assertRaises(ValueError, whale.get_transactions, 0, end_time='not_an_int', api_key='1234')

    def test_no_config_get_transactions_works_as_normal(self):
        whale = WhaleAlert()
        whale.transactions.get_transactions = mock.MagicMock().method()
        whale.transactions.get_transactions.return_value = (False, {}, {})
        expected = [mock.call(0, None, 'asdf', None, 500000, 100)]
        whale.get_transactions(0, api_key='asdf')
        self.assertEqual(whale.transactions.get_transactions.mock_calls, expected)


class FetchAndStoreData(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(TEST_WORKING_DIR)
        self.config = self.whale.get_configuration()
        self.status = self.whale.get_status()
        self.database = self.whale.get_database()

    def tearDown(self):
        cleanup_working_directories()

    def test_no_api_key_and_no_working_directory_returns_false_and_doesnt_call(self):
        whale = WhaleAlert()
        success = whale.fetch_and_store_data(0, 0)

        whale.transactions.get_transactions = mock.MagicMock().method()
        self.assertEqual(len(whale.transactions.get_transactions.mock_calls), 0)
        self.assertIs(success, False)

    def test_no_api_key_with_working_directory_calls_using_saved_api_key(self):
        self.config.set_value(settings.API_section_name, settings.API_option_private_key, 'zxcv')
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (True, good_transactions, good_status)

        success = self.whale.fetch_and_store_data(0)

        expected = [mock.call(0, None, 'zxcv', None, 500000, 100)]
        self.assertEqual(self.whale.transactions.get_transactions.mock_calls, expected)
        self.assertIs(success, True)

    def test_supplying_api_key_with_working_directory_calls_using_supplied_api_key(self):
        self.config.set_value(settings.API_section_name, settings.API_option_private_key, 'zxcv')
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (True, good_transactions, good_status)

        success = self.whale.fetch_and_store_data(0, api_key='poiu')

        expected = [mock.call(0, None, 'poiu', None, 500000, 100)]
        self.assertEqual(self.whale.transactions.get_transactions.mock_calls, expected)
        self.assertIs(success, True)

    def test_a_successful_call_causes_the_database_to_be_written(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (True, good_transactions, good_status)

        success = self.whale.fetch_and_store_data(0)
        database = self.whale.get_database()
        status = self.whale.get_status()

        successful_calls = status.get_value(settings.status_file_current_session_section_name,
                                            settings.status_file_option_successful_calls)
        failed_calls = status.get_value(settings.status_file_current_session_section_name,
                                        settings.status_file_option_failed_calls)

        df = database.table_to_df('bitcoin')
        self.assertEqual(len(df), 1)
        self.assertEqual(successful_calls, 1)
        self.assertEqual(failed_calls, 0)
        self.assertIs(success, True)

    def test_a_successful_call_with_failed_keys_causes_no_write_ands_a_failed_status(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (True, bad_transactions, good_status)

        success = self.whale.fetch_and_store_data(0)
        database = self.whale.get_database()
        status = self.whale.get_status()

        successful_calls = status.get_value(settings.status_file_current_session_section_name,
                                            settings.status_file_option_successful_calls)
        failed_calls = status.get_value(settings.status_file_current_session_section_name,
                                        settings.status_file_option_failed_calls)

        error_code = status.get_value(settings.status_file_last_failed_secion_name,
                                      settings.status_file_option_error_code)

        df = database.table_to_df('bitcoin')
        self.assertEqual(df.empty, True)
        self.assertEqual(successful_calls, 0)
        self.assertEqual(failed_calls, 1)
        self.assertIs(success, False)
        self.assertIs(error_code, 20)

    def test_a_failed_call_causes_no_write_to_database_and_failed_status(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (False, good_transactions, bad_status)

        success = self.whale.fetch_and_store_data(0)
        database = self.whale.get_database()
        status = self.whale.get_status()

        successful_calls = status.get_value(settings.status_file_current_session_section_name,
                                            settings.status_file_option_successful_calls)
        failed_calls = status.get_value(settings.status_file_current_session_section_name,
                                        settings.status_file_option_failed_calls)

        df = database.table_to_df('bitcoin')
        self.assertEqual(df, None)  # Nothing should have been written, so the table doesn't exist and none is returned
        self.assertEqual(successful_calls, 0)
        self.assertEqual(failed_calls, 1)
        self.assertIs(success, False)


class WritingCustomStatus(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(TEST_WORKING_DIR)
        self.config = self.whale.get_configuration()
        self.status = self.whale.get_status()
        self.database = self.whale.get_database()

    def tearDown(self):
        cleanup_working_directories()

    def test_writing_well_formated_status_returns_true(self):
        success = self.whale.write_custom_status(good_status)
        self.assertEqual(success, True)

    def test_writing_badly_formated_status_returns_false(self):
        success = self.whale.write_custom_status(bad_formatted_status)
        self.assertEqual(success, False)

    def test_writing_status_with_no_working_directory_returns_false(self):
        whale = WhaleAlert()
        success = whale.write_custom_status(good_status)
        self.assertEqual(success, False)


class WritingExcelFile(unittest.TestCase):
    def setUp(self):
        self.whale = WhaleAlert(TEST_WORKING_DIR)
        self.config = self.whale.get_configuration()
        self.status = self.whale.get_status()
        self.database = self.whale.get_database()

    def tearDown(self):
        cleanup_working_directories()

    def test_return_false_if_no_working_directory(self):
        whale = WhaleAlert()
        success = whale.to_excel()
        self.assertEqual(success, False)

    def test_return_false_no_databases_exist(self):
        success = self.whale.to_excel()
        self.assertEqual(success, False)

    def test_writes_an_excel_file(self):
        self.whale.transactions.get_transactions = mock.MagicMock().method()
        self.whale.transactions.get_transactions.return_value = (True, good_transactions, good_status)
        success = self.whale.fetch_and_store_data(0)

        success = self.whale.to_excel()
        exists = os.path.exists('whaleAlert.xlsx')
        self.assertEqual(exists, True)
        self.assertEqual(success, True)
        os.remove("whaleAlert.xlsx")
